/**
 * @file
 * @brief Inject constants from constants.hpp into QQbarThreshold.m
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.1
 *
 * @section LICENSE
 *
 * Copyright 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <iostream>
#include <fstream>
#include <map>
#include <sstream>
#include <stdexcept>

static const std::map<std::string, std::string> var_name{
  {"alpha_s_mZ", "alphaSmZDefault"},
  {"alpha_mZ", "alphamZ"},
  {"alpha_Y", "alphaY"},
  {"mu_alpha_Y", "muAlphaY"},
  {"mZ", "mZ"},
  {"mW", "mW"},
  {"m_Higgs", "mHiggsDefault"},
  {"alpha_QED", "alphaQED"},
  {"e_u", "eU"},
  {"e_d", "eD"},
  {"e", "eE"},
  {"G_F", "GF"},
  {"cw2", "cw2"},
  {"sw2", "sw2"},
  {"T3_nu", "T3Nu"},
  {"T3_e", "T3E"},
  {"T3_u", "T3U"},
  {"T3_d", "T3D"},
  {"mb_SI", "mbSI"},
  {"mu_thr", "muThr"},
  {"nl_bottom", "nlBottom"},
  {"nl_top", "nlTop"},
  {"mu_f_bottom", "mufBottom"},
  {"mu_f_top", "mufTop"},
  {"invGeV2_to_pb", "InvGeV2ToPb"},
};

std::string remove_multiline_comments(std::string in){
  for(
      size_t comment_begin = in.find("/*");
      comment_begin != std::string::npos;
      comment_begin = in.find("/*")
  ){
    size_t comment_end = in.find("*/", comment_begin + 2);
    if(comment_end == std::string::npos){
      throw std::runtime_error(
          "Comment has no end: " + in.substr(comment_begin, in.length())
      );
      comment_end += 2;
    }
    in.erase(comment_begin, comment_end - comment_begin);
  }
  return in;
}

std::string remove_singleline_comments(std::string in){
  for(
      size_t comment_begin = in.find("//");
      comment_begin != std::string::npos;
      comment_begin = in.find("//")
  ){
    size_t comment_end = in.find('\n', comment_begin + 2);
    if(comment_end == std::string::npos){
      throw std::runtime_error(
          "Comment has no end: " + in.substr(comment_begin, in.length())
      );
      comment_end += 1;
    }
    in.erase(comment_begin, comment_end - comment_begin);
  }
  return in;
}

std::string remove_comments(std::string in){
  return remove_singleline_comments(
      remove_multiline_comments(in)
  );
}

std::string slurp(std::ifstream & in){
  std::stringstream sstr;
  sstr << in.rdbuf();
  return sstr.str();
}

bool isword(char c){
  return std::isalnum(c) || c == '_';
}

bool is_part_of_word(size_t first, size_t size, std::string const & str){
  size_t last = first + size;
  return
    (first != 0 && isword(str[first - 1]))
    || (last != str.length() && isword(str[last]))
    ;
}

std::string replace_vars(std::string in){
  for(auto const & entry: var_name){
    std::string const & c_name = entry.first;
    std::string const & math_name = entry.second;
    for(
        size_t var_pos = in.find(c_name);
        var_pos != std::string::npos;
        var_pos = in.find(c_name, var_pos + 1)
    ){
      if(! is_part_of_word(var_pos, c_name.length(), in)){
        in.replace(var_pos, c_name.length(), math_name);
        var_pos += math_name.length();
      }
    }
  }
  return in;
}

void copy_constants(std::ifstream & const_in, std::ofstream & out){
  std::string const constants = remove_comments(slurp(const_in));
  for(auto const & entry: var_name){
    std::string const & c_name = entry.first;
    std::string const & math_name = entry.second;
    out << math_name;
    size_t var_pos = 0;
    for(bool found = false; !found;){
      var_pos = constants.find(c_name, var_pos);
      if(var_pos > 0 && !std::isspace(constants[var_pos - 1])){
        var_pos += c_name.length();
        continue;
      }
      if(var_pos == std::string::npos){
        throw std::runtime_error(
            "Variable " + c_name + "not found\n"
        );
      }
      var_pos += c_name.length();
      var_pos = constants.find_first_not_of(" \t\n\f\v\r", var_pos);
      if(constants[var_pos] == '='){
        found = true;
        size_t var_end = constants.find(';', var_pos);
        if(var_end == std::string::npos){
          throw std::runtime_error(
              "Definition of " + c_name  + "has no end\n"
          );
        }
        out << ' '
            << replace_vars(constants.substr(var_pos, var_end - var_pos))
            << ";\n";
      }
    }
  }
}

int main(int argc, char** argv){

  if(argc < 3){
    std::cout << "Usage: inject_constants constants_file math_package outfile\n";
    std::exit(1);
  }
  std::ifstream constants_in(argv[1]);
  if(! constants_in.is_open()){
    std::cerr << "Failed to open " << argv[1] << '\n';
    std::exit(1);
  }
  std::ifstream math_in(argv[2]);
  if(! math_in.is_open()){
    std::cerr << "Failed to open " << argv[2] << '\n';
    std::exit(1);
  }

  std::ofstream out(argv[3]);
  if(! out.is_open()){
    std::cerr << "Failed to open output file" << argv[3] << 'n';
    std::exit(1);
  }

  for(std::string line; std::getline(math_in, line);){
    static const std::string begin_private = "Begin[\"`Private`\"];";
    out << line << '\n';
    if(line.substr(0, begin_private.size()) == begin_private){
      copy_constants(constants_in, out);
#ifdef QQT_GRID_DIRECTORY
      out << "GridDirectory = \"" QQT_GRID_DIRECTORY "\";\n";
#endif
    }
  }
}
