/**
 * @file
 * @brief Mellin transforms of polylogarithms
 *
 * Mellin transforms of polylogarithms are required to express hypergeometric
 * functions in terms of nested harmonic sums.
 *
 * @see hypergeometric.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>

#include "complex.hpp"
#include "harmonic_sums.hpp"
#include "zeta.hpp"
#include "psi.hpp"

namespace hypergeometric{

  namespace detail{

    typedef std::complex<double> complex;
    using namespace complex;
    using namespace zeta;
    using harmonic_sums::harmonic_sum;
    using psi::psi;

    //! Mellin transforms of polylogarithms.
    /**
     * @tparam N   Order of the polylogarithm.
     * @param z    Argument of the Mellin transformed expression.
     * @returns    \f${\cal M}\bigg[\frac{{\rm Li}_N(1-t)}{1-t}\bigg](z)$\f,
     *             where \f${\cal M}$\f denotes the Mellin transform
     *             with respect to \f$ t $\f.
     *
     *  The Mellin transforms are implemented for integer order between 2 and 4.
     */
    template<int N>
    struct Mellin_Li{
      complex operator()(complex z);
    };

    template<>
    struct Mellin_Li<2>{
      complex operator()(complex z){
	return (
		- zeta_2*harmonic_sum<1>(z)
		+ harmonic_sum<1>(z)*harmonic_sum<2>(z)
		+ harmonic_sum<3>(z)
		- harmonic_sum<2, 1>(z)
		+ zeta_3
		);
      }
    };

    template<>
    struct Mellin_Li<3>{
      complex operator()(complex z){
	return (
		+ 1/2.*(
		       + pow(harmonic_sum<1>(z),2)*psi<1>(z+1)
		       + zeta_2*harmonic_sum<2>(z)
		       - harmonic_sum<4>(z)
		       )
		+ harmonic_sum<3, 1>(z)
		- harmonic_sum<2, 1, 1>(z)
		- harmonic_sum<1>(z)*(
					+ harmonic_sum<3>(z)
					- harmonic_sum<2, 1>(z)
					+ zeta_3
					)
		+ zeta_4
		);
      }
    };

    template<>
    struct Mellin_Li<4>{
      complex operator()(complex z){
	return (
		- harmonic_sum<3>(z)/18*(pi*pi - 15*harmonic_sum<2>(z))
		+ harmonic_sum<5>(z)
		+ harmonic_sum<2>(z)*harmonic_sum<2, 1>(z)/2
		- harmonic_sum<3, 2>(z)
		+ harmonic_sum<1>(z)*(
					- zeta_4
					- pi*pi/12*harmonic_sum<2>(z)
					+ harmonic_sum<4>(z)/2
					- harmonic_sum<3, 1>(z)
					+ harmonic_sum<2, 1, 1>(z)
					)
		- harmonic_sum<2, 2, 1>(z)
		+ harmonic_sum<3, 1, 1>(z)
		- harmonic_sum<2, 1, 1, 1>(z)
		- harmonic_sum<1>(z)*harmonic_sum<1>(z)*harmonic_sum<1>(z)*psi<1>(z+1)/6
		+ harmonic_sum<2>(z)*zeta_3/2
		+ harmonic_sum<1>(z)*harmonic_sum<1>(z)/2*(
					    + harmonic_sum<3>(z)
					    - harmonic_sum<2, 1>(z)
					    + zeta_3
					    )
		+ zeta_5
		);
      }
    };

  }
}
