/**
 * @file
 * @brief Wilson coefficients for the matching between PNRQCD, NRQCD and QCD
 *
 * This header contains the matching coefficients for the potentials and
 * the QCD vector current.
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.1
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include "parameters.hpp"
#include "QCD/colour.hpp"
#include "QCD/beta.hpp"
#include "zeta.hpp"

namespace QQbar_threshold{
  namespace Wilson_coefficients{
    using QCD::c_F;
    using QCD::c_A;
    using QCD::T_f;
    using QCD::d_abcd_AF;
    using QCD::d_abcd_FF;
    using QCD::T_f;
    using QCD::beta;
    using namespace zeta;

    /****************** Potential Coefficients *******************************
     *
     * The potential matching coefficients.
     * Definitions and references in
     * Beneke, Kiyo, Schuller, Phys.Lett.B658:222-229,2008, arXiv:0705.4518
     * Beneke, Kiyo, Schuller, arXiv:1312.4791 \cite Beneke:2013jia
     *
     ************************************************************************/

    //! Matching coefficients for the Coulomb potential
    /**
     * @tparam N   Loop order from 1 to 3.
     * @param nl   Number of active (light) flavours.
     *
     * References (third order):
     * Anzai, Kiyo, Sumino, Phys.Rev.Lett.104:112003,2010, arXiv:0911.4335
     * Smirnov, Smirnov, Steinhauser, Phys.Rev.Lett.104:112002,2010, arXiv:0911.4742
     *
     * Padea3 is the Pade estimate of a3 (now obsolete, kept for comparison with
     * older results).
     */
    template<int N>
    double a(int nl);

    //! Order ep^1 contributions to matching coefficients for the Coulomb potential
    /**
     * @tparam N   Loop order. Only N = 1 is required.
     * @param nl   Number of active (light) flavours.
     *
     * This correction is required for double insertions of the Coulomb
     * potential and a singular potential.
     */
    template<int N>
    double a_ep(int nl);

    template<>
    inline double a<1>(int nl){
      return 31./9.*c_A - 20./9.*nl*T_f;
    }

    template<>
    inline double a_ep<1>(int nl){
      return (188./27. - 11./6.*zeta_2)*c_A + (- 56./27. + zeta_2/3.)*nl;
    }

    template<>
    inline double a<2>(int nl){
      return 400./81.*pow(nl,2)*pow(T_f,2) - c_F*nl*T_f*(55./3. - 16*zeta_3) +
    pow(c_A,2)*(4343./162. + 4*pow(pi,2) - pow(pi,4)/4. + 22./3.*zeta_3) -
    c_A*nl*T_f*(1798./81. + 56./3.*zeta_3);

    }

    template<>
    inline double a<3>(int nl){
      static constexpr double c_FA = 15./16.;
      static constexpr double c_FF = 5./96.;
      return
	+ 502.24*pow(c_A,3) - 136.39*c_FA
	+ nl*(
	    -56.83*c_FF - 709.717*pow(c_A,2)*T_f
	    + pow(c_F,2)*T_f*(286./9. + 296./3.*zeta_3 - 160*zeta_5)
	    + c_A*c_F*T_f*(-71281./162. + 264*zeta_3 + 80*zeta_5)
	)
	+ pow(nl,2)*(
	    + c_F*pow(T_f,2)*(14002./81. - 416./3.*zeta_3)
	    + c_A*pow(T_f,2)*(12541./243. + 64./135.*pow(pi,4) + 368./3.*zeta_3)
	)
	- 8000./729.*pow(nl,3)*pow(T_f,3);
      // Padea3 =20358.4 - 4438.4*p.nl + 227.2*pow(p.nl,2);
    }

    //! NLO Coulomb potential including QED correction
    /**
     * The classical QED Coulomb potential has the same momentum dependence
     * form as the terms proportional to a1, a2,... in the QCD Coulomb
     * potential. Since alpha ~ alpha_s^2, it can be absorbed into an effective
     * a1.
     */
    inline double a1_eff(parameters const & p){
      return
	+ a<1>(p.nl)
	+ p.contributions.v_QED_Coulomb[0]*4*pi/c_F*pow(p.e_q/p.alpha_s,2)*p.alpha
	;
    }

    //! Matching coefficients for the 1/m (or 1/r^2) potential
    /**
     * @tparam N   Loop order from 1 to 2.
     * @param nl   Number of active (light) flavours.
     *

     * References (third order):
     * C. Anzai, Y. Kiyo, Y. Sumino, Phys.Rev.Lett.104:112003,2010, arXiv:0911.4335
     * A.V. Smirnov, V.A. Smirnov, M. Steinhauser, Phys.Rev.Lett.104:112002,2010,
     *  arXiv:0911.4742
     *
     * Padea3 is the Pade estimate of a3 (now obsolete, kept for comparison with
     * older results).
     */

    template<int N>
    double b(int nl);

    //! Order ep^1 contributions to  matching coefficients for the 1/m potential
    /**
     * @tparam N   Loop order. N = 1 and N = 2 are required.
     * @param nl   Number of active (light) flavours.
     *
     * b_ep<2> is taken from Beneke, Kiyo, Marquard, Penin, Piclum,
     * Seidel, Steinhauser, arXiv:1401.3005
     */
    template<int N>
    double b_ep(int nl);

    //! Order ep^2 contributions to  matching coefficients for the 1/m potential
    /**
     * @tparam N   Loop order. Only N = 1 is required.
     * @param nl   Number of active (light) flavours.
     */
    template<int N>
    double b_ep2(int nl);

    template<>
    inline double b<1>(int){
      return -c_A + c_F/2.;
    }

    template<>
    inline double b_ep<1>(int){
      return c_A + c_F*(-1 + log(2)) - 2*c_A*log(2);
    }

    template<>
    inline double b_ep2<1>(int){
      return (c_F*(5*pow(pi,2) + 12*(-4 + 2*log(2))*log(2)) -
     2*c_A*(5*pow(pi,2) + 12*(-2 + 2*log(2))*log(2)))/24.;
    }

    template<>
    inline double b<2>(int nl){
      return -101./36.*pow(c_A,2) + 65./18.*c_A*c_F + 49./36.*c_A*nl*T_f -
     2./9.*c_F*nl*T_f - 4./3.*pow(c_A,2)*log(2) - 8./3.*c_A*c_F*log(2);
    }

    template<>
    inline double b_ep<2>(int nl){
      return (
	  + c_F*c_A*(-631./108. - 15./16.*pow(pi,2) + 65./9. *log(2) - 8./3.*pow(log(2),2))
	  + pow(c_A,2)*(-1451./216. - 161./72.*pow(pi,2) - 101./18.*log(2) - 4./3.*pow(log(2),2))
	  + c_A*T_f*nl*(115./54. + 5./18.*pow(pi,2) + 49./18.*log(2))
	  + c_F*T_f*nl*(17./27. - 11./36.*pow(pi,2) - 4./9.*log(2))
      );
    }

    //! Leading-order coefficient for the 1/m^2 (or delta) potential
    constexpr double v0(int){
      return -2./3.;
    }

    //! Order ep contribution to v0
    constexpr double v0_ep(int){
      return -5./18.;
    }

    //! Order ep^2 contribution to v0
    constexpr double v0_ep2(int){
      return 4./27.;
    }

    /***************************************************************************
     *
     * vq1,vm1: one-loop corrections to the coefficients of the 1/m^2 potential
     * vp1: part of one-loop correction to the momentum-dependent potential
     * as defined in
     * Beneke, Kiyo, Schuller, arXiv:1312.4791 \cite Beneke:2013jia
     *
     **************************************************************************/

    constexpr double vq1(int nl){
      return -11./27.*c_A - c_F/3. + 40./27.*nl*T_f;
    }

    constexpr double vq1_ep(int nl){
      return c_F*(2 - 7./36.*pi*pi) + c_A*(-419./81. + 77./216.*pi*pi) +
	nl*(274./81. - 2./27.*pi*pi)*T_f;
    }

    constexpr double vm1(int){
      return -29./9.*c_A - c_F/3. + 4./15.*T_f;
    }

    constexpr double vm1_ep(int){
      return c_F*(-10 + pi*pi/36.) + c_A*(379./54. + pi*pi/24.);
    }

    constexpr double vp1(int nl){
      return 31./9.*c_A - 20./9.*nl*T_f;
    }

    constexpr double vp1_ep(int nl){
      return c_A*(188./27. - 19./36.*pi*pi) + nl*(-112./27. + pi*pi/9.)*T_f;
    }

    /************** Tree-level coefficients for production operators ****************
     *
     * See e.g. eqs. (3), (4) in \ref Beneke:2010mp
     *
     *******************************************************************************/

    //! Coefficient of vector production operator
    inline
    double Cv_p(parameters const & p){
      if(p.production == production_channel::photon_only) return e*p.e_q;
      double const v_q = p.c_q[detail::vector];
      static const double v_e = detail::c_e<detail::vector>();
      return e*p.e_q + v_q*v_e*p.s/(p.s - mZ*mZ);
    }

    //! Coefficient of axialvector production operator
    inline
    double Ca_p(parameters const & p){
      if(p.production == production_channel::photon_only) return 0;
      double const v_q = p.c_q[detail::vector];
      static const double a_e = detail::c_e<detail::axial>();
      return -v_q*a_e*p.s/(p.s - mZ*mZ);
    }


    /***************************** Current Coefficients *****************************

     Matching coefficients for the QCD vector current. The expansion
     is organized as follows

     cv = 1+cv<1> alpha_s/(4 pi)+cv<2> (alpha_s/(4 pi))^2+cv<3> (alpha_s/(4 pi))^3+...

     and similar for dv and ca (p-wave matching).

     References:

     c2 from:
     M. Beneke, A. Signer, V.A. Smirnov, Phys.Rev.Lett.80:2535-2538 (1998),
     hep-ph/9712302;
     A. Czarnecki, K. Melnikov, Phys.Rev.Lett.80:2531-2534 (1998),
     hep-ph/9712222

     c3 from:
     P. Marquard, J.H. Piclum, D. Seidel, M. Steinhauser,
     Nucl.Phys.B758:144 (2006), hep-ph/0607168;
     Phys.Lett.B678:269-275(2009), arXiv:0904.0920v1[hep-ph].
     arXiv:1401.3004

     The singlet contributions to c3 are not yet known.
     THE UNKNOWN TERMS ARE TEMPORARILY SET TO ZERO.

     Higgs contribution to cv from Eiras, Steinhauser NPB 757 (2006) 197

     d1 from M.E. Luke, M.J. Savage, Phys.Rev.D57:413 (1998), hep-ph/9707313

    ************************************************************************************/

    template<int N>
    double cv_QCD(parameters const & p);

    template<int N>
    double cv_Higgs(parameters const & p);

    template<>
    inline double cv_QCD<1>(parameters const &){
      return - 8*c_F;
    }

    inline double cv_2_constant(parameters const & p){
      return
 	+ 16*c_F*(
	  + c_F*(23./8.-zeta_3/2.-79./36.*pow(pi,2)+pow(pi,2)*log(2))
	  + c_A*(-151./72.-13./4.*zeta_3+89./144.*pow(pi,2)-5./6.*pow(pi,2)*log(2))
	  + T_f*p.nl*(11./18.)
	  + T_f*(22./9.-2./9.*pow(pi,2))
      );
    }

    template<>
    inline double cv_QCD<2>(parameters const & p){
      return
	+ 2*log(p.mu/p.m)*(
	    + 4*beta(0, p.nl)*cv_QCD<1>(p)
	    + c_F*pow(pi,2)*(-8./3.*c_F-4*c_A)
	)
	+cv_2_constant(p);
    }

    template<>
    inline double cv_QCD<3>(parameters const & p){
      return
	+(
	    + pow(4*beta(0, p.nl),2)*pow(2*log(p.mu/p.m), 2)
	    + 16*beta(1, p.nl)*2*log(p.mu/p.m)
	)*cv_QCD<1>(p)
	+ 2*4*beta(0, p.nl)*2*log(p.mu/p.m)*(
	    + c_F*pow(pi,2)*2*log(p.mu/p.m)*(-8./3.*c_F-4*c_A)
	    + cv_2_constant(p)
	)
	+ pow(2*log(p.mu/p.m), 2)*c_F*pow(pi,2)/4.*(
	    pow(c_F,2)*(-40)+
	    c_F*c_A*(-188./9.)+
	    pow(c_A,2)*(128./3.)+
	    c_F*T_f*p.nl*(-128./9.)+
	    c_A*T_f*p.nl*(-64./3.)
	)
	+ 2*log(p.mu/p.m)*c_F*pow(pi,2)/2.*(
	    pow(c_F,2)*(-72+192*log(2))+
	    c_F*c_A*(-1888./27.-96*log(2))+
	    pow(c_A,2)*(-256./9.-96*log(2))+
	    c_F*T_f*p.nl*(800./27.)+
	    c_A*T_f*p.nl*(296./9.)+
	    c_F*T_f*(-32./5.)
	)
	+ 64*c_F*T_f*p.nl*(
	    c_F*(46.692)+
	    c_A*(39.623)+
	    T_f*(-557./162.+26./81.*pow(pi,2))+
	    T_f*p.nl*(-163./162.-4./27.*pow(pi,2))
	)
	+ 64*c_F*T_f*(
	    c_F*(-0.846)+
	    c_A*(-0.1)+
	    T_f*(-427./162.+158./2835.*pow(pi,2)+16./9.*zeta_3)
	)
	+ pow(4,3)*(-2090.) // gluonic part from arXiv:1401.3004
	;
    }

    template<int N>
    double cv(parameters const & p){
      return
	+ p.contributions.cv[N-1] * pow(p.alpha_s/(4*pi),N) * cv_QCD<N>(p)
	+ p.contributions.cv_Higgs[N-2] * p.Yukawa_t2/(2*pi*pi) * cv_Higgs<N>(p)
	;
    }

    template<>
    inline double cv<1>(parameters const & p){
      return p.contributions.cv[0]*p.alpha_s/(4*pi)*cv_QCD<1>(p);
    }

    template<int N>
    double ca(parameters const & p);

    template<>
    inline double ca<1>(parameters const & p){
      return p.contributions.ca[0]*p.alpha_s/(4*pi)*-4*c_F;
    }

    template<>
    inline double cv_Higgs<2>(parameters const & p){
      double const yH_sqr = pow(p.m/p.m_Higgs,2);
      double arctan = (yH_sqr > 1./4.)?
	sqrt(4*yH_sqr - 1)*atan(sqrt(4*yH_sqr - 1)):
	sqrt(1 - 4*yH_sqr)*atanh(sqrt(1 - 4*yH_sqr))
	;

      return
	(3*yH_sqr - 1)/(12*yH_sqr)
	- (2 + yH_sqr*(- 9 + 12*yH_sqr))/(48*pow(yH_sqr,2))*log(yH_sqr)
	- (-2 + yH_sqr*( 5 - 6*yH_sqr))/(24*yH_sqr)*arctan/yH_sqr
	;
    }

    template<>
    inline double cv_Higgs<3>(parameters const & p){
      double const yH1b = 1 - pow(p.m/p.m_Higgs,2);
      return p.alpha_s/pi*c_F*(
	  3./2.*zeta_2*(1 - yH1b)*log(p.m/p.mu)
	  -5.76 + yH1b*(5.533 + yH1b*(-0.171 + yH1b*(0.0124 + yH1b*(0.0304 + 0.0296*yH1b))))
      );
    }

    /*********** subleading hard matching coefficient **********/

    template<int N>
    double dv_QCD(parameters const & p);

    template<>
    inline double dv_QCD<1>(parameters const & p){
      return -16*c_F*(1./3. + 2*log(p.mu/p.m));
    }

    template<int N>
    double dv(parameters const & p){
      return p.contributions.dv[N-1] * pow(p.alpha_s/(4*pi),N) * dv_QCD<N>(p);
    }

    template<>
    inline double dv<0>(parameters const & p){
      return  p.contributions.dv[0];
    }

    template<>
    inline double dv<1>(parameters const & p){
      return  p.contributions.dv[1]*p.alpha_s/(4*pi)*dv_QCD<1>(p);
    }

    /*********** electroweak matching coefficients **********/
    template<int N>
    complex Cv_EW(parameters const & p);

    template<int N>
    complex Ca_EW(parameters const & p);

  }
}
