/**
 * @file
 * @brief Potential insertion functions
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#pragma once

#include <complex>

#include "insertion_sums.hpp"

namespace QQbar_threshold{
  namespace detail{

    typedef std::complex<double> complex;

    //! Single insertions
    /**
     * @tparam I     Insertion type.
     * @param lam    The value of \f$\lambda = \frac{\alpha_s C_F}{2} \sqrt{-\frac{m}{E + i\Gamma}}$\f.
     * @param p      Further parameters.
     * @return       The insertion of a single potential into the Green function
     *
     * The insertion type is usually the exponent of 1/q^2 in the potential,
     * e.g. insertion<i1+a1*ep>(lam, p) corresponds to the insertion of
     * V = 1/(q^2)^(i1+a1*ep).
     *
     * An exception is insertion<delta>(lam, p), which denotes an insertion of
     * the d3-dimensional delta potential (in momentum space)
     * V_delta = (2pi)^d3*delta^d3(p-p')
     */
    template<ins_type I>
    complex insertion(complex lam, parameters const & p);

    //! Double insertions
    template<ins_type I1, ins_type I2>
    complex insertion(complex lam, parameters const & p);

    //! Triple insertions
    template<ins_type I1, ins_type I2, ins_type I3>
    complex insertion(complex lam, parameters const & p);

    //! TODO: documentation
    template<ins_type I>
    complex insertion_1(complex lam, parameters const & p);

    //! Single insertions of singular potential
    /**
     *  For singular potentials we obtain contributions from higher
     *  orders in ep of the coefficient. Thus
     *  insertion_RJ<I>(lam, p)
     *  corresponds to the insertion of the potential
     *  V = ep^J*1/(q^2)^I
     */
    template<ins_type I>
    complex insertion_R0(complex lam, parameters const & p);
    //! Insertion of V = ep*1/(q^2)^I
    template<ins_type I>
    complex insertion_R1(complex lam, parameters const & p);
    //! Insertion of V = ep^2*1/(q^2)^I
    template<ins_type I>
    complex insertion_R2(complex lam, parameters const & p);

  }
}
