/**
 * @file
 * @brief Implementation of function to load grids
 *
 * @see load_grid.hpp
 *
 * @author  Andreas Maier <a.maier@tum.de>
 * @version 1.0.1
 *
 * @section LICENSE
 *
 * Copyright 2014, 2015, 2016 Andreas Maier
 *
 * This file is part of the QQbar_threshold library.
 *
 * The QQbar_threshold library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of
 * the License, or (at your option) any later version.
 *
 * The QQbar_threshold library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the QQbar_threshold library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 */
#include <string>
#include <fstream>
#include <array>
#include <vector>
#include <complex>
#include <cassert>
#include <cstdlib>
#include <tuple>
#include <utility>
#include <exception>
#include <stdexcept>
#include <limits>

#include "load_grid.hpp"
#include "insertion_sums.hpp"
#include "interpolation_grid.hpp"
#include "non-resonant.hpp"
#include "word_exp.hpp"

namespace QQbar_threshold{
  typedef std::complex<double> complex;

  std::string grid_directory(){
    static constexpr char dir[] = QQT_GRID_DIRECTORY;
    return dir;
  }

  namespace detail{
    //! grid for stored values of insertion sums
    std::array<
      interpolation_grid<complex>,
      last_insertion_sum + 1
      > insertion_grids;


    //! grid for integrals in non-resonant constribution
    std::array<interpolation_grid<double>, last_integral_type+1>
      h_interpolations{
        #include "h_grids.dat"
      };

    double h_interpolation(integral_type  t, double x, double yw){
      const double xmin = h_interpolations[t].x_min();
      const double xmax = h_interpolations[t].x_max();
      if((x < xmin)|| (x >= xmax)){
        throw std::out_of_range(
            "loaded non-resonant grid too small:\n"
            "ratio mw^2/m^2 = " + std::to_string(x)
            + " not in range ["
            + std::to_string(xmin) + ", " + std::to_string(xmax)
            +")"
        );
      }
      return h_interpolations[t](x, yw);
    }
  };

  void load_grid(char const* filename){
    using word_exp::word_exp;
    std::ifstream in;
    in.open(to_string(word_exp(filename)));
    if(! in.is_open()){
      throw std::ios_base::failure(
    	  std::string{"Failed to open '"} + filename + "'"
      );
    }
    load_grid(std::move(in));
  }

  void load_grid(std::string const & filename){
    load_grid(filename.c_str());
  }

  void load_nonresonant_grid(char const* filename){
    using word_exp::word_exp;
    std::ifstream in;
    in.open(to_string(word_exp(filename)));
    if(! in.is_open()){
      throw std::ios_base::failure(
    	  std::string{"Failed to open '"} + filename + "'"
      );
    }
    load_nonresonant_grid(std::move(in));
  }

  void load_nonresonant_grid(std::string const & filename){
    load_nonresonant_grid(filename.c_str());
  }

  //! Remove mathematica-specific notation
  void sanitize(std::string & str){
    for(
	size_t pos = str.find("*^");
	pos != std::string::npos;
	pos = str.find("*^", pos)
    ){
      str.replace(pos,2,"e");
    }
    for(
	size_t pos = str.find("*I");
	pos != std::string::npos;
	pos = str.find("*I", pos)
    ){
      str.erase(pos,2);
    }
    for(
	size_t pos = str.find('\t');
	pos != std::string::npos;
	pos = str.find('\t', pos)
    ){
      str[pos] = '\n';
    }
  }

  template<typename T>
  using table_t = std::vector< std::vector<T> >;

  //! Get number of different (rescaled) width values in a grid
  /**
   * Not very cache friendly, but there are usually not many
   * different widths
   */
  size_t count_width_points(table_t<complex> const & grid){
    if(grid.empty()) return 0;
    double last_width = imag(grid[0][0]);
    for(size_t i = 1; i < grid.size(); ++i){
      assert(imag(grid[i][0]) < 0);
      if(imag(grid[i][0]) > last_width) return i;
      last_width = imag(grid[i][0]);
    }
    return grid.size();
  }

  template<typename T>
  T to_value(const char *str, char **str_end);

  template<>
  double to_value<double>(const char *str, char **str_end){
    return std::strtod(str, str_end);
  }

  template<>
  complex to_value<complex>(const char *str, char **str_end){
    complex res;
    res.real(std::strtod(str, str_end));
    const char * im_str = *str_end;
    if((*im_str != '+') && (*im_str != '-')) return res;
    res.imag(std::strtod(im_str, str_end));
    return res;
  }

  //! Read data from tsv file into a table
  template<typename T>
  table_t<T> read_tsv(std::ifstream & in){
    table_t<T> values;
    std::string line;
    while(std::getline(in, line)){
      //ignore comments
      if(line.front() == '#') continue;
      //extract values
      sanitize(line);
      values.emplace_back();
      char const * pos = line.c_str();
      char * end;
      values.back().emplace_back(to_value<T>(pos, &end));
      do{
        pos = end;
        values.back().emplace_back(to_value<T>(pos, &end));
      } while(*end);
    }
    return values;
  }

  void load_grid(std::ifstream && in){
    using namespace detail;

    // read table from file
    auto grid = read_tsv<complex>(in);

    // get dimensions (width values) x (energy values)
    size_t num_width_points = count_width_points(grid);
    assert(grid.size() % num_width_points == 0);
    size_t num_energy_points = grid.size() / num_width_points;

    //re-sort into ascending order
    std::reverse(begin(grid), end(grid));

    // fill points into global interpolation grids
    std::vector<double> rescaled_width_points(num_width_points);
    for(size_t i = 0; i < num_width_points; ++i){
      rescaled_width_points[i] = imag(grid[i][0]);
    }

    std::vector<double> rescaled_energy_points;
    rescaled_energy_points.reserve(num_energy_points);
    for(size_t i = 0; i < grid.size(); i += num_width_points){
      rescaled_energy_points.emplace_back(real(grid[i][0]));
    }
    assert(rescaled_energy_points.size() == num_energy_points);

    std::array<
      matrix<complex>,
      last_insertion_sum + 1
      > insertion_sums;
    for(size_t ins = 0; ins <= last_insertion_sum; ++ins){
      insertion_sums[ins] = matrix<complex>(num_width_points, num_energy_points);
    }

    for(size_t i = 0; i < grid.size(); ++i){
      assert(grid[i].size() == last_insertion_sum + 2);
      size_t row = i % num_width_points;
      size_t col = i / num_width_points;
      for(size_t ins = 0; ins <= last_insertion_sum; ++ins){
        insertion_sums[ins](row, col) = grid[i][ins+1];
      }
    }

    for(size_t ins = 0; ins <= last_insertion_sum; ++ins){
      insertion_grids[ins] = interpolation_grid<complex>{
        rescaled_width_points,
        rescaled_energy_points,
        std::move(insertion_sums[ins])
      };
    }
  }

  struct coordinate_less{
    template<typename T>
    bool operator()(T a, T b){
      return std::tie(a[0], a[1]) < std::tie(b[0], b[1]);
    }
  };

  struct y_less{
    template<typename T>
    bool operator()(T a, T b){
      return a[1] < b[1];
    }
  };

  // TODO: code duplication with load_grid
  void load_nonresonant_grid(std::ifstream && in){
    using namespace detail;

    // read table from file
    auto grid = read_tsv<double>(in);
#ifndef NDEBUG
    for(auto const & line: grid){
      assert(line.size() == last_integral_type + 3);
    }
#endif
    std::sort(begin(grid), end(grid), coordinate_less{});

    // get coordinates
    std::vector<double> x, y;
    auto y_block_end = std::is_sorted_until(
        begin(grid), end(grid), y_less{}
    );
    for(auto it = begin(grid); it != y_block_end; ++it){
      y.emplace_back((*it)[1]);
    }
    assert(grid.size() % y.size() == 0);
    for(size_t i = 0; i < grid.size(); i+= y.size()){
      x.emplace_back(grid[i][0]);
    }

    std::array<
      matrix<double>,
      last_integral_type + 1
      > h_vals;
    for(size_t int_type = 0; int_type <= last_integral_type; ++int_type){
      h_vals[int_type] = matrix<double>(x.size(), y.size());
    }

    for(size_t i = 0; i < grid.size(); ++i){
      size_t row = i / y.size();
      size_t col = i % y.size();

      for(size_t int_type = 0; int_type <= last_integral_type; ++int_type){
        h_vals[int_type](row, col) = grid[i][int_type+2];
      }
    }

    for(size_t int_type = 0; int_type <= last_integral_type; ++int_type){
      h_interpolations[int_type] = interpolation_grid<double>{
        x, y,
        std::move(h_vals[int_type])
      };
    }

  }

  Et_Gammat_range grid_range(){
    using detail::insertion_grids;
    assert(! insertion_grids.empty());
    if(insertion_grids[0].empty()) return {0., 0., 0., 0.};
    return {
      insertion_grids[0].y_min(),
      insertion_grids[0].y_max(),
      insertion_grids[0].x_min(),
      insertion_grids[0].x_max()
    };
  }

}
